#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import unittest

from neptune_tensorboard.sync.internal.path_parser import parse_path_to_experiment_name, parse_path_to_hostname


class PathParserTest(unittest.TestCase):

    def test_parsing_experiment_name(self):
        # GIVEN
        test_cases = [
            (
                "events.out.tfevents.1551714927.jan-kowalski-pascal05-test-4vgt2",
                "untitled-tensorboard"
            ),
            (
                "deepsense/events.out.tfevents.1551714927.jan-kowalski-pascal05-test-4vgt2",
                "deepsense"
            ),
            (
                "deepsense/events.out.tfevents.1551719327.adam-timing-signal-benchmark11-worker-0",
                "deepsense"
            ),
            (
                "deepsense/events.out.tfevents.1552057986.adam-sem-prevents-overfit-hypothesis-05-worker-0",
                "deepsense"
            ),
            (
                "deepsense/events.out.tfevents.1552057986 (1).adam-sem-prevents-overfit-hypothesis-05-worker-0",
                "deepsense"
            ),
            (
                "mnist/logs/mnist_with_summaries/test/events.out.tfevents.1551288692.adam-nowak",
                "mnist/logs/mnist_with_summaries/test"
            ),
            (
                "mnist/logs/mnist_with_summaries/train/events.out.tfevents.1551288692.piotr-adamski",
                "mnist/logs/mnist_with_summaries/train"
            ),
            (
                "mnist/something",
                "mnist"
            ),
            (
                "mnist",
                "untitled-tensorboard"
            )
        ]

        for test_case in test_cases:
            # WHEN
            parsed_name = parse_path_to_experiment_name(test_case[0])

            # THEN
            self.assertEqual(parsed_name, test_case[1], "{} should equal {}".format(parsed_name, test_case[1]))

    def test_parsing_hostname(self):
        # GIVEN
        test_cases = [
            (
                "events.out.tfevents.1551714927.jan-kowalski-pascal05-test-4vgt2",
                "jan-kowalski-pascal05-test-4vgt2"
            ),
            (
                "deepsense/events.out.tfevents.1551714927.jan-kowalski-pascal05-test-4vgt2",
                "jan-kowalski-pascal05-test-4vgt2"
            ),
            (
                "deepsense/events.out.tfevents.1551719327.adam-timing-signal-benchmark11-worker-0",
                "adam-timing-signal-benchmark11-worker-0"
            ),
            (
                "deepsense/events.out.tfevents.1552057986.adam-sem-prevents-overfit-hypothesis-05-worker-0",
                "adam-sem-prevents-overfit-hypothesis-05-worker-0"
            ),
            (
                "deepsense/events.out.tfevents.1552057986 (1).adam-sem-prevents-overfit-hypothesis-05-worker-0",
                "adam-sem-prevents-overfit-hypothesis-05-worker-0"
            ),
            (
                "mnist/logs/mnist_with_summaries/test/events.out.tfevents.1551288692.adam-nowak",
                "adam-nowak"
            ),
            (
                "mnist/logs/mnist_with_summaries/train/events.out.tfevents.1551288692.piotr-adamski",
                "piotr-adamski"
            ),
            (
                "mnist/something",
                None
            ),
            (
                "mnist/data/events.out.tfevents.piotr-adamski",
                None
            )
        ]

        for test_case in test_cases:
            # WHEN
            parsed_hostname = parse_path_to_hostname(test_case[0])

            # THEN
            self.assertEqual(parsed_hostname, test_case[1], "{} should equal {}".format(parsed_hostname, test_case[1]))



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os

_EVENTS_FILE_PREFIX = "events.out.tfevents."


def parse_path_to_experiment_name(path):
    """
    Parses a relative path to tensorflow event file to a reasonable neptune-compatible name

    Arguments:
        path(str): a path relative to entrypoint directory

    Returns: a string representing sanitized project name based on the path, or "untitled-tensorboard" if name cannot
        be determined
    """
    experiment_name = os.path.dirname(path)
    if experiment_name:
        return experiment_name
    else:
        return "untitled-tensorboard"


def parse_path_to_hostname(path):
    """
    Parses a relative path to tensorflow event file to a hostname

    Arguments:
        path(str): a path relative to entrypoint directory

    Returns: a hostname or None if the file name did not match tensorflow events file
    """
    file_name = os.path.basename(path)
    if file_name.startswith(_EVENTS_FILE_PREFIX):
        timestamp_and_hostname = file_name[len(_EVENTS_FILE_PREFIX):]
        separator_index = timestamp_and_hostname.find('.')
        if separator_index >= 0:
            return timestamp_and_hostname[(separator_index + 1):]
        else:
            return None
    else:
        return None



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from neptune.exceptions import LibraryNotInstalled, NeptuneException

_integrated_with_keras = False


def integrate_with_keras(experiment_getter):
    global _integrated_with_keras  # pylint:disable=global-statement

    if _integrated_with_keras:
        return

    _integrate_with_keras(experiment_getter)

    _integrated_with_keras = True


def _integrate_with_keras(experiment_getter):
    try:
        import keras
    except ImportError:
        raise LibraryNotInstalled('keras')

    from keras.callbacks import BaseLogger, Callback  # pylint:disable=import-error

    class NeptuneLogger(Callback):

        def __init__(self, experiment=None, experiment_getter=None):
            super(NeptuneLogger, self).__init__()

            def get_exp():
                return experiment

            self._experiment_holder = get_exp if experiment is not None else experiment_getter

        def on_batch_end(self, batch, logs=None):  # pylint:disable=unused-argument

            if logs is None:
                return

            for metric, value in logs.items():

                try:
                    if metric in ('batch', 'size'):
                        continue
                    name = 'keras_on_batch_end_' + metric
                    self._experiment_holder().send_metric(channel_name=name,
                                                          x=value,
                                                          y=None)
                except NeptuneException:
                    pass

        def on_epoch_end(self, epoch, logs=None):  # pylint:disable=unused-argument

            if logs is None:
                return

            for metric, value in logs.items():
                try:
                    if metric in ('epoch', 'size'):
                        continue

                    name = 'keras_on_epoch_end_' + metric
                    self._experiment_holder().send_metric(channel_name=name,
                                                          x=value,
                                                          y=None)
                except NeptuneException:
                    pass

    class KerasAggregateCallback(Callback):

        def __init__(self, *callbacks):
            super(KerasAggregateCallback, self).__init__()
            self.callbacks = callbacks

        def set_params(self, params):
            for callback in self.callbacks:
                callback.params = params

        def set_model(self, model):
            for callback in self.callbacks:
                callback.model = model

        def on_epoch_begin(self, epoch, logs=None):
            for callback in self.callbacks:
                callback.on_epoch_begin(epoch, logs=logs)

        def on_batch_end(self, batch, logs=None):
            for callback in self.callbacks:
                callback.on_batch_end(batch, logs=logs)

        def on_epoch_end(self, epoch, logs=None):
            for callback in self.callbacks:
                callback.on_epoch_end(epoch, logs=logs)

    def monkey_patched_BaseLogger(*args, **kwargs):
        return KerasAggregateCallback(BaseLogger(*args, **kwargs),
                                      NeptuneLogger(experiment_getter=experiment_getter))

    keras.callbacks.BaseLogger = monkey_patched_BaseLogger



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import unittest

import mock
import tensorflow as tf
from bunch import Bunch

from neptune_tensorboard.sync.tensorflow_data_sync import TensorflowDataSync


class TensorflowDataSyncTest(unittest.TestCase):
    _EXPERIMENT = Bunch()
    _EXPERIMENT.limits = {
        'channels': {
            'numeric': 100,
            'text': 10,
            'image': 10
        }
    }

    def test_limiting_simple_value_summaries(self):
        # GIVEN
        summaries = []
        for i in range(1000):
            summaries.append(self.__create_summary(
                self.__create_simple_value('num', i)
            ))
        # AND
        tag_buckets = {}

        # WHEN
        step = 0
        for summary in summaries:
            # pylint: disable=protected-access
            TensorflowDataSync._apply_limit(self._EXPERIMENT, step, summary, tag_buckets)
            step += 1

        # THEN
        self.assertEqual(len(tag_buckets['num']), 100)
        self.assertEqual(tag_buckets['num'][0].value.simple_value, 0)
        self.assertEqual(tag_buckets['num'][0].x, 0)
        self.assertEqual(tag_buckets['num'][99].value.simple_value, 999)
        self.assertEqual(tag_buckets['num'][99].x, 999)

    def test_limiting_image_summaries(self):
        # GIVEN
        summaries = []
        for i in range(100):
            summaries.append(self.__create_summary(
                self.__create_image('image', "imageContent{}".format(i))
            ))
        # AND
        tag_buckets = {}

        # WHEN
        step = 0
        for summary in summaries:
            # pylint: disable=protected-access
            TensorflowDataSync._apply_limit(self._EXPERIMENT, step, summary, tag_buckets)
            step += 1

        # THEN
        # pylint: disable=line-too-long
        self.assertEqual([i.value.image for i in tag_buckets['image']],
                         ["imageContent{}".format(i) for i in range(90, 100)])
        self.assertEqual([i.x for i in tag_buckets['image']], list(range(90, 100)))

    def test_limiting_text_summaries(self):
        # GIVEN
        summaries = []
        for i in range(100):
            summaries.append(self.__create_summary(
                self.__create_text('text', "[text{}]".format(i))
            ))
        # AND
        tag_buckets = {}

        # WHEN
        step = 0
        for summary in summaries:
            # pylint: disable=protected-access
            TensorflowDataSync._apply_limit(self._EXPERIMENT, step, summary, tag_buckets)
            step += 1

        # THEN
        self.assertEqual([i.value.tensor.string_val for i in tag_buckets['text']],
                         ["[text{}]".format(i) for i in range(90, 100)])
        self.assertEqual([i.x for i in tag_buckets['text']], list(range(90, 100)))

    @staticmethod
    def __create_summary(value):
        summary = mock.MagicMock()
        summary.value = [value]
        return summary

    @staticmethod
    def __create_simple_value(tag, simple_value):
        value = mock.MagicMock()
        value.tag = tag
        value.WhichOneof.return_value = 'simple_value'
        value.simple_value = simple_value
        return value

    @staticmethod
    def __create_image(tag, content):
        value = mock.MagicMock()
        value.tag = tag
        value.WhichOneof.return_value = 'image'
        value.image = content
        return value

    @staticmethod
    def __create_text(tag, text):
        value = mock.MagicMock()
        value.tag = tag
        value.WhichOneof.return_value = 'tensor'
        value.tensor = mock.MagicMock()
        value.tensor.dtype = tf.string
        value.tensor.string_val = text
        return value



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import print_function

import os
import random
import re
import traceback

import click
import tensorflow as tf
from future.moves import collections, sys

from neptune_tensorboard.integration.tensorflow_integration import TensorflowIntegrator
from neptune_tensorboard.sync.internal.path_parser import parse_path_to_experiment_name, parse_path_to_hostname


class TensorflowDataSync(object):
    _RECORD = collections.namedtuple('Record', 'x value')

    def __init__(self, project, path):
        self._project = project
        self._path = path

    def run(self):
        for root, _, run_files in os.walk(self._path):
            for run_file in run_files:
                try:
                    self._load_single_run(os.path.join(root, run_file))
                except Exception as e:
                    print("Cannot load run from file '{}'. ".format(run_file) + str(e), file=sys.stderr)
                    try:
                        traceback.print_exc(e)
                    except: # pylint: disable=bare-except
                        pass

    def _load_single_run(self, path):
        click.echo("Loading {}...".format(path))
        run_path = os.path.relpath(path, self._path)
        run_id = re.sub(r'[^0-9A-Za-z_\-]', '_', run_path).lower()
        exp_name = parse_path_to_experiment_name(run_path)
        hostname = parse_path_to_hostname(run_path)
        if not self._experiment_exists(run_id, exp_name):
            with self._project.create_experiment(name=exp_name,
                                                 properties={
                                                     'tf/run/path': run_path
                                                 },
                                                 tags=[run_id],
                                                 upload_source_files=[],
                                                 abort_callback=lambda *args: None,
                                                 upload_stdout=False,
                                                 upload_stderr=False,
                                                 send_hardware_metrics=False,
                                                 run_monitoring_thread=False,
                                                 handle_uncaught_exceptions=True,
                                                 hostname=hostname or None) as exp:
                tf_integrator = TensorflowIntegrator(lambda *args: exp)
                self._load_single_file(exp, path, tf_integrator)
            click.echo("{} was saved as {}".format(run_path, exp.id))
        else:
            click.echo("{} is already synced".format(run_path))

    def _experiment_exists(self, run_id, run_path):
        existing_experiments = self._project.get_experiments(tag=run_id)
        return any(exp.name == run_path and exp.state == 'succeeded' for exp in existing_experiments)

    @staticmethod
    def _load_single_file(experiment, path, tf_integrator):
        tag_buckets = {}
        for record in tf.train.summary_iterator(path):
            if record.graph_def:
                graph_def = tf.GraphDef()
                graph_def.ParseFromString(record.graph_def)
                tf_integrator.add_graph_def(graph_def, path)
            TensorflowDataSync._apply_limit(experiment, record.step, record.summary, tag_buckets)

        for tag in tag_buckets:
            for record in tag_buckets[tag]:
                tf_integrator.add_value(record.x, record.value)

    @staticmethod
    def _apply_limit(experiment, step, summary, tag_buckets):
        for value in summary.value:
            if value.tag not in tag_buckets:
                tag_buckets[value.tag] = []

            bucket = tag_buckets[value.tag]
            bucket.append(TensorflowDataSync._RECORD(step, value))

            field = value.WhichOneof('value')
            if field == 'simple_value':
                channel_type = 'numeric'
            elif field == 'image':
                channel_type = 'image'
            elif field == 'tensor' and value.tensor.dtype == tf.string:
                channel_type = 'text'
            else:
                continue

            if len(bucket) > experiment.limits['channels'][channel_type]:
                if channel_type == 'numeric':
                    del bucket[random.randint(1, len(bucket) - 2)]
                else:
                    bucket.pop(0)



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import unicode_literals

import io
import os
import time
import uuid

import tensorflow as tf
from PIL import Image
from future.builtins import object, str
from tensorflow.core.framework import summary_pb2  # pylint:disable=no-name-in-module
from neptune.exceptions import NeptuneException

_integrated_with_tensorflow = False


def integrate_with_tensorflow(experiment_getter):
    global _integrated_with_tensorflow  # pylint:disable=global-statement

    if _integrated_with_tensorflow:
        return
    _integrate_with_tensorflow(experiment_getter)
    _integrated_with_tensorflow = True


class TensorflowIntegrator(object):

    def __init__(self, experiment_getter=None):

        self._experiment_holder = experiment_getter
        self._summary_writer_to_graph_id = {}

    def add_summary(self, summary, global_step=None):

        if isinstance(summary, bytes):
            summ = summary_pb2.Summary()
            summ.ParseFromString(summary)
            summary = summ

        x = self._calculate_x_value(global_step)
        for value in summary.value:
            try:
                self.add_value(x, value)
            except NeptuneException:
                pass

    def add_value(self, x, value):
        field = value.WhichOneof('value')

        if field == 'simple_value':
            self._send_numeric_value(value.tag, x, value.simple_value)
        elif field == 'image':
            self._send_image(value.tag, x, value.image.encoded_image_string)
        elif field == 'tensor' and value.tensor.dtype == tf.string:
            string_values = []
            for _ in range(0, len(value.tensor.string_val)):
                string_value = value.tensor.string_val.pop()
                string_values.append(string_value.decode("utf-8"))
                self._send_text(value.tag, x, ", ".join(string_values))

    def add_graph_def(self, graph_def, log_dir):
        writer = self.get_writer_name(log_dir)
        if writer in list(self._summary_writer_to_graph_id.keys()):
            graph_id = self._summary_writer_to_graph_id[writer]
        else:
            graph_id = str(uuid.uuid4())
            self._summary_writer_to_graph_id[writer] = graph_id
        try:
            self._experiment_holder().send_graph(graph_id, str(graph_def))
        except NeptuneException:
            pass

    def _send_numeric_value(self, value_tag, x, simple_value):
        self._experiment_holder().send_metric(channel_name=value_tag,
                                              x=x,
                                              y=simple_value)

    def _send_image(self, image_tag, x, encoded_image):
        image_desc = "({}. Step {})".format(image_tag, x)
        self._experiment_holder().send_image(channel_name=image_tag,
                                             x=x,
                                             y=Image.open(io.BytesIO(encoded_image)),
                                             name=image_desc,
                                             description=image_desc)

    def _send_text(self, value_tag, x, text):
        self._experiment_holder().send_text(channel_name=value_tag,
                                            x=x,
                                            y=text)

    @staticmethod
    def get_writer_name(log_dir):
        return os.path.basename(os.path.normpath(log_dir))

    @staticmethod
    def _calculate_x_value(global_step):
        if global_step is not None:
            return int(global_step)
        else:
            return time.time()


def _integrate_with_tensorflow(experiment_getter):
    tensorflow_integrator = TensorflowIntegrator(experiment_getter=experiment_getter)

    # pylint: disable=no-member, protected-access, no-name-in-module, import-error

    _add_summary_method = tf.summary.FileWriter.add_summary
    _add_graph_def_method = tf.summary.FileWriter._add_graph_def

    def _neptune_add_summary(summary_writer, summary, global_step=None):
        tensorflow_integrator.add_summary(summary, global_step)
        _add_summary_method(summary_writer, summary, global_step=global_step)

    def _neptune_add_graph_def(summary_writer, graph_def, global_step=None):

        try:
            from tf.tensorboard.backend import process_graph
        except ImportError:
            from tensorboard.backend import process_graph

        # Restricting graph only to UI relevant information.
        process_graph.prepare_graph_for_ui(graph_def)
        # pylint: disable=protected-access
        tensorflow_integrator.add_graph_def(graph_def, summary_writer.get_logdir())
        _add_graph_def_method(summary_writer, graph_def, global_step=global_step)

    tf.summary.FileWriter.add_summary = _neptune_add_summary
    tf.summary.FileWriter._add_graph_def = _neptune_add_graph_def

    return tensorflow_integrator
